/*
 *Copyright (c) 2000-2002 Viola Systems Ltd.
 *All rights reserved.
 *
 *Redistribution and use in source and binary forms, with or without 
 *modification, are permitted provided that the following conditions 
 *are met:
 *
 *1. Redistributions of source code must retain the above copyright 
 *notice, this list of conditions and the following disclaimer.
 *
 *2. Redistributions in binary form must reproduce the above copyright 
 *notice, this list of conditions and the following disclaimer in the 
 *documentation and/or other materials provided with the distribution.
 *
 *3. The end-user documentation included with the redistribution, if 
 *any, must include the following acknowledgment:
 *	"This product includes software developed by Viola 
 *	Systems (http://www.violasystems.com/)."
 *
 *Alternately, this acknowledgment may appear in the software itself, 
 *if and wherever such third-party acknowledgments normally appear.
 *
 *4. The names "OpenTCP" and "Viola Systems" must not be used to 
 *endorse or promote products derived from this software without prior 
 *written permission. For written permission, please contact 
 *opentcp@opentcp.org.
 *
 *5. Products derived from this software may not be called "OpenTCP", 
 *nor may "OpenTCP" appear in their name, without prior written 
 *permission of the Viola Systems Ltd.
 *
 *THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED 
 *WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 *MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 *IN NO EVENT SHALL VIOLA SYSTEMS LTD. OR ITS CONTRIBUTORS BE LIABLE 
 *FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 *CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 *SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
 *BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 *OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 *EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *====================================================================
 *
 *OpenTCP is the unified open source TCP/IP stack available on a series 
 *of 8/16-bit microcontrollers, please see <http://www.opentcp.org>.
 *
 *For more information on how to network-enable your devices, or how to 
 *obtain commercial technical support for OpenTCP, please see 
 *<http://www.violasystems.com/>.
 */

/** \file ethernet.c
 *	\brief OpenTCP Ethernet protocol and driver for Realtek's RTL8019AS
 *	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *		\li Vladan Jovanovic (vladan.jovanovic@violasystems.com)
 *	\version 1.0
 *	\date 19.2.2002
 *	\bug
 *	\warning
 *	\todo
 *		\li How to adopt to different MCU speeds in the code?
 *		\li NE2000DMAInit ??? What is it used for?
 *  
 *	OpenTCP implementation of Ethernet protocol and driver 
 */
#include <inet/system.h>

/************************************************/
/* Modified by Renesas Technology America, Inc. */
/* Makers of M16C, H8, SH                       */
/* http://www.renesas.com                       */
/************************************************/

#include "smc91111.h"

/************ Configurations ************/

#if M16C_62P
#define BASE_ADDR 0x10300
#endif

#if M32C_85
#define BASE_ADDR 0x100300
#endif

#if M32C_83
#define BASE_ADDR 0x100300
#endif

#define POLLED_ETHERNET 0
#define INTERRUPT_ETHERNET 1

#define RX_BUF_SIZE 1550

/* For our transmit buffer, use the same size that we specified in system.h */
#define TX_BUF_SIZE NETWORK_TX_BUFFER_SIZE

static void delay(int ms_delay);
static void SMsC_Write(unsigned int, unsigned int);
static unsigned int SMsC_Read(unsigned int);
static UINT16 PhyAccess(UINT8 write, UINT8 addr, UINT16 data);
void SmscIntr(void);

#define SMSC_WRITE(offset,data)  *((volatile FAR UINT16 *)(BASE_ADDR + offset) ) = data
#define SMSC_READ(offset)        *((volatile FAR UINT16 *)(BASE_ADDR + offset))
//#define SMSC_WRITE SMsC_Write
//#define SMSC_READ SMsC_Read


/* For PhyAccess() function */
#define PHY_WRITE 1
#define PHY_READ 0


#define BANK_SELECT(b)            SMSC_WRITE(BSR_REG, b )

UINT8	NE2000NextPktPtr;			/**< Start address of next packet in the Ethernet controller */
UINT8 	NE2000CurrPktPtr;			/**< Start address of current packet in the Ethernet controller */

UINT8 	EtherSleep = 0;	/**< Used for storing state of Ethernet controller (0 = awake; 1 = sleeping) */

UINT8 rx_buffer_filled = 0;	/* used to signal when a Rx interrupt has occured and the buffer has been filled */


/* allocate tx/rx buffers */
UINT8 eth_pkt_rx_buf[RX_BUF_SIZE];
UINT8 * eth_rx_buf_ptr;
UINT8 eth_pkt_tx_buf[TX_BUF_SIZE];
UINT8 * eth_tx_buf_ptr;

struct _ETH_INFO
{
	UINT32 total_rx_packets;
	UINT32 total_tx_packets;
	UINT8 full_duplex:1;
	UINT8 speed_100:1;
	UINT8 linked:1;
} eth_info = {0,0,0,0,0};

/** \brief Used for storing various information about the received Ethernet frame
 *	
 *	Fields from Ethernet packet (dest/source hardware address, 
 *	protocol, frame size, start of the Ethernet packet in Ethernet controller)
 *	are stored in this structure. These values are later used from upper
 *	layer protocols (IP, ARP). See ethernet_frame definition for more
 *	information about struct fields.
 *
 */	
struct ethernet_frame received_frame;

/** \brief Used for storing various information about the Ethernet frame that will be sent
 *	
 *	Fields from Ethernet packet (dest/source hardware address, 
 *	protocol, frame size) are stored in this structure by the
 *	upper layer protocols (IP, ARP, other). These values are 
 *	then used for initializing transmission of an Ethernet frame.
 *	See ethernet_frame definition for more
 *	information about struct fields.
 *
 */
struct ethernet_frame send_frame;

/* Functions not used in this implementation */
//void outNE2000 (UINT8 reg , UINT8 dat) { }	
//UINT8 inNE2000 (UINT8 reg) { }
//void NE2000CheckOverFlow (void) { }
//void NE2000DMAInit (UINT8 page) { }
//void NE2000EnterSleep (void) { }
//void NE2000ExitSleep (void) { }


/** \brief Write data to the same NE2000 register as before
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\param dat new register content
 *
 *	Invoke this function to write data to the same register as 
 *	in previously invoked outNE2000(). Saves some cycles because
 *	data direction registers and address are not set up.
 *
 *	Use SEND_NETWORK_B() macro instead of invoking this function
 *	directly.
 */
void outNE2000again (UINT8 dat)
{
	/* Actually, because this function wants to write a byte and our */
	/* Ethernet controller only wants to write words, we will write
	/* every thing to a buffer, then send it later */
	*eth_tx_buf_ptr++ = dat;
}
/** \brief Write buffer data to the same NE2000 register as before
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 20.07.2003
 *	\param buf pointer to buffer from which we're taking data
 *	\param len number of bytes from buffer to write
 *
 *	Use SEND_NETWORK_BUF() macro instead of invoking this function
 *	directly.
 */
void outNE2000againbuf (UINT8* buf, UINT16 len)
{
	while(len--)
	{
		*eth_tx_buf_ptr++ = *buf++;
	}
}

/** \brief Continue reading byte(s) from NE2000
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\return Register content
 *
 *	Invoke this function to continue reading contents of a NE2000 register,
 *	usefull when reading a sequence of bytes from Ethernet controller.
 *
 *	Use RECEIVE_NETWORK_B() macro instead of directly invoking this
 *	function for easier driver change.
 */
UINT8 inNE2000again (void)
{
	/* Actually, because this function wants to read a byte and our */
	/* Ethernet controller only wants to read words, we will read
	/* every thing from a buffer that we filled already */
	return 	*eth_rx_buf_ptr++;
}
/** \brief Read bytes from NE2000 to buffer
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 20.07.2003
 *
 *	Invoke this function to continue reading contents of a NE2000 register,
 *	only this time to a buffer directly.
 *
 *	Use RECEIVE_NETWORK_BUF() macro instead of directly invoking this
 *	function for easier driver change.
 */
void inNE2000againbuf (UINT8* buf, UINT16 len)
{
	while(len--)
	{
		*buf++ = *eth_rx_buf_ptr++;
	}
}


/** \brief Check to see if new frame has been received
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\return
 *		\li #TRUE - if there is a new packet waiting
 *		\li #FALSE - if there is no new packet waiting
 *
 *	Invoke this function to check if there is a new Ethernet frame
 *	in the NIC's buffer waiting to be processed.
 */
UINT8 NE2000CheckRxFrame (void)
{
	int UINT16 status;

	/* Turn Receive on if it's not already enabled */
	BANK_SELECT(0);
	SMSC_WRITE(RCR_REG,RCR_STRIP_CRC|RCR_RXEN);		/* strip CRC from received packets, enable receiving */

	BANK_SELECT(2);
	/* Read Interrupt Status Register to determine if a new packet has arrived */
	status = SMSC_READ(IST_REG);

	/* Check if a receive error has occured */
	if( status & IST_RX_OVRN_INT)
	{
		/* This bit will stay latched untill reset */
		SMSC_WRITE(ACK_REG,IST_RX_OVRN_INT);
//		SMSC_WRITE(MMUCR_REG,MMUCR_RESET);
//		while ( SMSC_READ(MMUCR_REG) & MMUCR_BUSY );		/* wait for MMU to become not busy */
//		return(FALSE);
	}

	if(status & IST_RCV_INT)
	{
		/* A new packet has arrived */
		return(TRUE);
	}
	return(FALSE);
}

/** \brief Discard current received frame
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *
 *	Discard the current received frame.
 */
void NE2000DumpRxFrame (void)
{
	rx_buffer_filled = 0;	// allow the next packet to come in
}
 
 
/** \brief Initialize and configure Ethernet
 *	\ingroup core_initializer
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\param mac Pointer to NIC's hardware address
 *
 *	Invoke this function at startup to properly initialize NIC's registers,
 *	HW address and operation.
 */ 
void NE2000Init (UINT8* mac)
{
	UINT16 tmp,tmp2;

	// for interrupt
	volatile char junk;

	/* Reset Chip */
	BANK_SELECT(0);						/* Select Bank 0 */
	SMSC_WRITE(RCR_REG,RCR_SOFTRST);	/* software reset ethernet chip */
	for(tmp=0;tmp<20;tmp++)
		delay(10);
	SMSC_WRITE(RCR_REG,0);				/* drive reset bit low again */


	BANK_SELECT(0);						/* Select Bank 0 */
//	SMSC_WRITE(RPCR_REG, RPCR_ANEG | 0x18 );		/* Auto-Negotiate the PHY conneciton, LEDB is RX recv */
	SMSC_WRITE(RPCR_REG, RPCR_ANEG | 0xF8 );		/* Auto-Negotiate the PHY conneciton, LEDB is Rx, LEDA is Tx */

	BANK_SELECT(1);						/* Select Bank 1 */
#if (POLLED_ETHERNET)
	SMSC_WRITE(CTR_REG, CTR_AUTO_RELEASE);		/* set AUTO_RELEASE */
#else
//	SMSC_WRITE(CTR_REG, CTR_AUTO_RELEASE | CTR_LE_ENABLE | CTR_CR_ENABLE | CTR_TE_ENABLE ); /* set AUTO_RELEASE and enable error interrupts  */
	SMSC_WRITE(CTR_REG, CTR_AUTO_RELEASE ); /* set AUTO_RELEASE and enable error interrupts  */
#endif

	/* make sure we can read what we just wrote */
	tmp = SMSC_READ(CTR_REG);
	if( !(tmp & CTR_AUTO_RELEASE) )	/* AUTO_RELEASE set? */
		while(1);		/* ERROR - Stop here */


#if (0)
	/** EEPROM PRESENT ** Include this section if you have an EPROM that
	/* initilizes the chip at boot-up.
	/* read MAC address and write it to eth_src_hw_addr. This assumes the
	MAC address was stored in EEPROM and already loaded into the chip */
	*((short *)(&mac[4])) = SMSC_READ(IAR1_IAR0);
	*((short *)(&mac[2])) = SMSC_READ(IAR3_IAR2);
	*((short *)(&mac[0])) = SMSC_READ(IAR5_IAR4);
#else
	/* ADDRESS 0 corresponds to the first address byte on the cable */
	/* OpenTCP defines the localHW[6] array as [5]:[4]:[3]:[2]:[1]:[0] */
	tmp = *(mac++);
	tmp = (tmp << 8) + *(mac++);
	SMSC_WRITE(IAR5_IAR4, tmp);
	tmp = *(mac++);
	tmp = (tmp << 8) + *(mac++);
	SMSC_WRITE(IAR3_IAR2, tmp);
	tmp = *(mac++);
	tmp = (tmp << 8) + *(mac++);
	SMSC_WRITE(IAR1_IAR0, tmp);
#endif

	BANK_SELECT(2);
	SMSC_WRITE(MMUCR_REG,MMUCR_RESET);				/* Write RESET MMU command */
	while ( SMSC_READ(MMUCR_REG) & MMUCR_BUSY );	/* Wait for MMU to become not busy */

	BANK_SELECT(0);
	SMSC_WRITE(RCR_REG,RCR_STRIP_CRC | RCR_RXEN);	/* strip CRC from received packets, enable receiving */
	SMSC_WRITE(TCR_REG,TCR_PAD_EN | TCR_TXENA);		/* turn on hardware padding & enable xmit */

#if (INTERRUPT_ETHERNET)
	/* enable interrupts */
	BANK_SELECT(2);
	tmp = SMSC_READ(MSK_REG);
//	tmp |= (MSK_RCV_INT | MSK_RX_OVRN_INT);
	tmp |= MSK_RCV_INT;
	SMSC_WRITE(MSK_REG,tmp);

	/* enable interrupts for M16C/62P */
	int2ic = 0x14;		/* enable INT2 pin, rising edge triggered, priority level 4 */	

//	P_INTC.IER.BIT.IRQ0E = 0;		/* disable IRQ0 interrupt */
//	junk = P_INTC.ISR.BIT.IRQ0F;
//	P_INTC.ISR.BIT.IRQ0F = 0;		/* clear IRQ0 interrupt */
//	P_INTC.ISCR.BIT.IRQ0SC = 2;		/* interrupt on rising edge */
//	P_INTC.ITSR.BIT.ITS0 = 1;		/* select P80 as input pin */
//	P_INTC.IPRA.BIT.UU = 4;			/* set priority to 4 */
//	P_INTC.IER.BIT.IRQ0E = 1;		/* enable IRQ0 interrupt */
#endif


	/* RESET THE PHY LAYER */
	PhyAccess(PHY_WRITE,PHY_CNTL_REG,PHY_CNTL_RST);		/* RESET */
	do
	{
		delay(50);
	}
	while( PhyAccess(PHY_READ,PHY_CNTL_REG,0) & PHY_CNTL_RST );	/* wait for RESET bit to clear */

	/* Need a long delay here because something is not finished initilizing yet and will not work if*/
	/* the phy is turned on too soon (don't know what) */
	delay(500);

	/* turn on phy layer and let link */
	PhyAccess(PHY_WRITE,PHY_CNTL_REG,PHY_CNTL_ANEG_EN);	

	DisplayString(LCD_LINE1,"Waiting ");
	DisplayString(LCD_LINE2,"Eth Link");

	do
	{
		tmp = PhyAccess(PHY_READ,PHY_STAT_REG,0);	/* READ PHY Status REGISTER 1 */
	}
	while(!(tmp & PHY_STAT_LINK) );

	eth_info.linked = 1;

	/* Keep reading INT register to make sure all bits have been set */
	do
	{
		tmp2 = tmp;
		tmp = PhyAccess(PHY_READ,PHY_INT_REG,0);	/* READ PHY Status Output (and Interrupt status) Register */
	}
	while( tmp != tmp2);


	/* Record Speed */
	if( tmp & PHY_INT_SPDDET)
	{
		eth_info.speed_100 = 1;
		DisplayString(LCD_LINE1,"100 Mbps");
	}
	else
	{
		DisplayString(LCD_LINE1,"10 Mbps ");	
	}

	/* Record Duplex */
	if( tmp & PHY_INT_DPLXDET)
	{
		eth_info.full_duplex = 1;
		DisplayString(LCD_LINE2,"Full DPX");
	}
	else
	{
		DisplayString(LCD_LINE2,"Half DPX");	
	}	

	delay(2000);			/* 2 sec delay to read message */		

	/* Set RXEN bit to enable receiving */

	/* If porting to the SMC91C11x add code to detect and setup PHY here */

}

/** \brief Checks if new Ethernet frame exists and initializes variables
 *		accordingly
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasytems.com)
 *	\date 19.02.2002
 *	\return
 *		\li #TRUE - new frame exists, variables initialized so further
 *			processing may start
 *		\li #FALSE - no new Ethernet frame detected
 *
 *	Invoke this function periodically (better not directly but by invoking
 *	NETWORK_CHECK_IF_RECEIVED() macro) to test for newly-arrived Ethernet
 *	packets.
 *
 *	Once this function detects a new Etherent packet, internal variables 
 *	are initialized (NE2000CurrPktPtr, NE2000NextPktPtr and received_frame
 *	structure) and TRUE is returned, indicating that processing of
 *	newly-arrived frame may begin.
 */
UINT8 NE2000ReceiveFrame (void)
{
	UINT16 packet_size;
	UINT16 tmp;

#if (INTERRUPT_ETHERNET)
	/* Turn Receive on if it's not already enabled */
	BANK_SELECT(0);
	SMSC_WRITE(RCR_REG,RCR_STRIP_CRC|RCR_RXEN);		/* strip CRC from received packets, enable receiving */
	
	if(rx_buffer_filled == 0)
		if( NE2000CheckRxFrame() == FALSE )
			return(FALSE);		/* no new packets yet */
	if(rx_buffer_filled == 1)
		return(TRUE);		/* buffer was filled by interrupt routine */

	/* if this far, then interrupt routine wants the buffer to get filled */
	rx_buffer_filled = 0;		// reset flag in case we get error filling buffer
#endif

#if (POLLED_ETHERNET)
	if( NE2000CheckRxFrame() == FALSE ) 
 		return(FALSE);
#endif

// 	BANK_SELECT(0);
//	SMSC_WRITE(RCR_REG,0);		/* Disable Receiveing packets. This helps in debug because if you hit
//							   a breakpoint, all your memory in the Ethernet chip will fill up and you
//							   won't be able to allocate memory to transmit. */

	/* A new packet has arrived so read it out */
	BANK_SELECT(2);								/* Select Bank 2 */
	SMSC_WRITE(PTR_REG,PTR_RCV|PTR_AUTOINC|PTR_READ);		/* Set up the POINTER REGISTER 
									   Set RCV (to signify the RX FIFO)
									   Set AUTO_INCR (to signify incrementing the fifo pointer)
									   Set READ (to signify reading out data)
									   Pointer index at 0 (start at begining) */

	/* First word in FIFO is status */
	tmp = SMSC_READ(DATA_REG);

	/* Make sure there were no errors in the packet */
	if (tmp & (RS_ALGNERR|RS_BADCRC|RS_TOOSHORT|RS_TOOLONG) )
		return FALSE;

	/* Next word in FIFO is count (take off upper 5 reserved bits)*/
	/* This number will always be even */
	packet_size = 0x07FF & SMSC_READ(DATA_REG);
	/* subtract 6 bytes for status, byte count, and control word. */
	packet_size -= 6;

	received_frame.frame_size = packet_size;		/* save full size of frame */
	packet_size -= ETH_HEADER_LEN;					/* Ethernet header will not be stored in buffer */

	/* Check if the packet data will fit into our receive buffer. */
	if( packet_size > RX_BUF_SIZE	)
	{
		/* packet too big for recv buffer */
		/* remove and free this packet */
		while ( SMSC_READ(MMUCR_REG) & MMUCR_BUSY );		/* wait for MMU to become not busy */
		SMSC_WRITE(MMUCR_REG,MMUCR_RELEASE);				/* remove and release top of Rx FIFO */
		eth_info.total_rx_packets++;	/* still count it as an incoming packet */
	 	return(FALSE);					/* no packet received */
 	}

/* NOT USED
 	NE2000CurrPktPtr = 0;
 	NE2000NextPktPtr = 0;
*/
 	/* Record destination Ethernet Address	*/
	tmp = SMSC_READ(DATA_REG);
	received_frame.destination[5] = (UINT8) tmp;	
	received_frame.destination[4] = (UINT8) (tmp >> 8);	
	tmp = SMSC_READ(DATA_REG);
	received_frame.destination[3] = (UINT8) tmp;	
	received_frame.destination[2] = (UINT8) (tmp >> 8);	
	tmp = SMSC_READ(DATA_REG);
	received_frame.destination[1] = (UINT8) tmp;	
	received_frame.destination[0] = (UINT8) (tmp >> 8);	
	
 	/* Record senders Ethernet address 		*/
	tmp = SMSC_READ(DATA_REG);
	received_frame.source[5] = (UINT8) tmp;	
	received_frame.source[4] = (UINT8) (tmp >> 8);	
	tmp = SMSC_READ(DATA_REG);
	received_frame.source[3] = (UINT8) tmp;	
	received_frame.source[2] = (UINT8) (tmp >> 8);	
	tmp = SMSC_READ(DATA_REG);
	received_frame.source[1] = (UINT8) tmp;	
	received_frame.source[0] = (UINT8) (tmp >> 8);

 	/* Record Protocol (account for ethernet chip being little endian) */
	tmp = SMSC_READ(DATA_REG);
 	received_frame.protocol = (tmp>>8) + (tmp<<8);

	/* Now read in the rest of the data to our receive buffer */
	eth_rx_buf_ptr = eth_pkt_rx_buf;
	tmp = packet_size;

	while(tmp)
	{
		* (UINT16*)eth_rx_buf_ptr = SMSC_READ(DATA_REG);
		eth_rx_buf_ptr +=2;
		tmp -=2;
	}

	/* Read the last word which will contain the control byte to determine the number of bytes were odd */
	tmp = SMSC_READ(DATA_REG);
	/* If the frame size was odd, then the last byte was the lower part of the word we read */
	if( tmp & 0x2000 )
	{
		*eth_rx_buf_ptr++ = (UINT8) tmp;
		received_frame.frame_size++;		/* Remember to adjust the size of the data */
	}

	/* Reset pointer to the begining of the buffer */
	eth_rx_buf_ptr = eth_pkt_rx_buf;

	/* Free up that memory */
	while ( SMSC_READ(MMUCR_REG) & MMUCR_BUSY );		/* wait for MMU to become not busy */
	SMSC_WRITE(MMUCR_REG,MMUCR_RELEASE);				/* remove and release top of Rx FIFO */

 	/* Give the next layer data start buffer index from the start	*/
 	received_frame.buf_index = 0;
 	
	eth_info.total_rx_packets++;

	rx_buffer_filled = 1; 	/* signal that the buffer has been filled */

 	ETH_DEBUGOUT("Ethernet Frame Received\n\r");

 	return(TRUE);						/* Indicate we got packet */

}

/** \brief Initialize transmission of new packet
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\param page Address in Ethernet controller where outgoing packet
 *		buffered
 *
 *	This function is used to initialize transmission of an Ethernet 
 *	packet. Packet is created starting from <i>page</i> address.
 *
 *	Do not invoke this function directly, but instead use 
 *	NETWORK_SEND_INITIALIZE() macro.
 */
void InitTransmission (UINT8 page)
{
	/* Simply set up pointer for transmit buffer used by outNE2000again() */
	eth_tx_buf_ptr = eth_pkt_tx_buf;
}

/** \brief Write Ethernet Header to transmission buffer
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\param frame Pointer to ethernet_frame structure containing information
 *		about the new Ethernet frame
 *
 *	Invoke this function (through NETWORK_ADD_DATALINK() macro) to create
 *	an Ethernet header in the NIC's transmit buffer. Do this only after
 *	invoking NETWORK_SEND_INITIALIZE() macro.
 *
 */
void NE2000WriteEthernetHeader (struct ethernet_frame* frame)
{
	INT8 i;
	
	/* Write destination Ethernet Address	*/
	
	for(i=ETH_ADDRESS_LEN-1; i >= 0; i--) {
		outNE2000again(frame->destination[i]);
	}

	/* Write sender (our) Ethernet address	*/
	
	for(i=ETH_ADDRESS_LEN-1; i >= 0; i--) {
		outNE2000again(frame->source[i]);
	}
	
	/* Write protocol						*/
	outNE2000again( (UINT8)(frame->protocol >> 8) );
	outNE2000again( (UINT8)frame->protocol );

}

/** \brief Initialize reading from NE2000
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\param pos Position from the start of current frame from which DMA
 *		is initialized.
 *
 *	Invoke this function (through NETWORK_RECEIVE_INITIALIZE() macro) 
 *	to prepair NIC for reading from the specified position.
 */
void NE2000DMAInit_position (UINT16 pos)
{
	/* Simply set up pointer for receive buffer used by inNE2000again() */
	eth_rx_buf_ptr = eth_pkt_rx_buf + pos;
}

/** \brief Instruct NIC to send the Ethernet frame
 * 	\author 
 *		\li Jari Lahti (jari.lahti@violasystems.com)
 *	\date 19.02.2002
 *	\param len Length of packet stored in the NIC's buffer
 *	\warning
 *		\li Minimum Ethernet packet size is 64 bytes
 *
 *	Invoke this function (through NETWORK_COMPLETE_SEND() macro) when
 *	the whole packet is formed inside NIC's memory and is ready to be 
 *	sent. Proper length of the packet must be supplied so that NIC
 *	knows how much data to put on the line.
 */
void NE2000SendFrame ( UINT16 len )
{
 	/* After filling Tx buffer, call this function to send it */
 	/* Input: 	Len: packet length							  */

	UINT16 temp,num_pages,txpktno;

 	/* Check if we need to insert Pad bytes */
 	while(len < 50)
	{
 		outNE2000again((BYTE)0x00);
 		len++;
 	} 	
 	
	/* add Ethernet header */
 	len += ETH_HEADER_LEN;

#if (POLLED_ETHERNET)
	BANK_SELECT(0);
	SMSC_WRITE(TCR_REG,SMSC_READ(TCR_REG) | TCR_TXENA);    /* Set TXENA bit to enable transmits. This bit may have 
												  been cleared by an error with the last packet, so we
												  have to make sure it's set */
#endif

	/* Allocate transmit memory in the Ethernet chip */
	/* Add 6 bytes for status word, byte count and control word */
	/* Determine the number of 256 byte pages that will be needed */
	num_pages = (len+6) / 256;		// NOTE: a value of 0 = 1 page, 1 = 2 pages, etc...

	BANK_SELECT(2);
	while(SMSC_READ(MMUCR_REG) & MMUCR_BUSY);			// Wait for BUSY bit to be clear
	SMSC_WRITE(MMUCR_REG,(MMUCR_ALLOC | num_pages));	/* Write ALLOCATE TX MEM command along with the number of
												   pages needed */ 
	/* Sit in a loop and wait for space to get allocated. If we sit too long, then 
	   we know that the incoming receive packets have used up all the memory inside
	   the Ethernet chip, so we need to start taking some of them out */
	temp = 20;		// How many times we should check before we flush the Ethernet chip
	while(1)
	{
		/* Keep checking the Interrupt Status Reg for the ALLOC INT bit to go high */
		if ( SMSC_READ(IST_REG) & IST_ALLOC_INT )
		{
			/* Find out what the packet number is of the memory we just allocated */
			txpktno = SMSC_READ(ARR_REG);
			
			/* Check if allocation failed by checking the FAILED bit */
			if( txpktno & ARR_FAILED )
				return;		/* ERROR - Allocation failed */

			/* Change the Packet Number register to the packet we just allocated */
			SMSC_WRITE(PNR_REG,txpktno);
			break;
		}

		/* After after 20 tries, we don't get an allocation, start deleting incoming packets */
		if(--temp == 0)
		{
			while(SMSC_READ(MMUCR_REG) & MMUCR_BUSY);		// Wait for BUSY bit to be clear
			SMSC_WRITE(MMUCR_REG,MMUCR_RELEASE);				/* Remove and relase top of Rx FIFO */
			temp = 20;
		}
	}

	SMSC_WRITE(PTR_REG,PTR_AUTOINC);	/* Set AUTO INCR. (to inc FIFO pointer automatically)
								   Clear RCV (to signify Tx FIFO
								   Clear READ (to signify write)
								   Celar ETEN (no Early trans underrun detection)
								   Clear AutoTx (we will manually enqueue this packet)
								   Pointer value is 0 (meaning we will start at the very begining of the FIFO) */

	/* Start writting data in */
	SMSC_WRITE(DATA_REG,0x0000);        /* write status word */

	/* Now write the data out to the Ethernet chip */
	eth_tx_buf_ptr = eth_pkt_tx_buf;
	temp = len & 0xFFFE;				/* force to even number */

	/* Write in byte count (must always be even) */
	/* Add 6 for status word, byte count and control word */
	SMSC_WRITE(DATA_REG,temp+6);

	while(temp)
	{
		// write a word at a time
		SMSC_WRITE(DATA_REG, *(UINT16 *)eth_tx_buf_ptr);
		eth_tx_buf_ptr += 2;
		temp -= 2;
	}
	/* Now write the control word. If there is 1 left byte to send, set the */
	/* ODD bit to indicate that the bottom word is the last byte */
	if(len & 1)
	{
		/* odd number of data */
		SMSC_WRITE(DATA_REG, *eth_tx_buf_ptr | 0x2000 );	/* Set ODD bit in control word */
	}
	else
	{
		/* even number of data */
		SMSC_WRITE(DATA_REG,0);		/* Write the control word */
	}


	/* Now write queue the packet for tranmitting by writting to the MMU Command
	   Register. We have to do this manually because we are not using Auto Transmit
	   The packet number to be enqueued is taken from the PNR register which we haven't
	   changed yet. */
	SMSC_WRITE(MMUCR_REG,MMUCR_ENQUEUE);	/* Write ENQUEUE PACKET NUMBER INTO TX FIFO command */

	BANK_SELECT(0);
	/* Keeps checking transmit status register for the result of the transmit */
	while(1)
	{
		temp = SMSC_READ(EPHSR_REG);
		if( temp & EPHSR_TX_SUC)
		{
			/* Transmit successful */
			/* Because we set the AutoRelease bit in the CTR register, the memory for
			   packet we just transmitted is automatically released because TX was successful */
			/* Can clear TXEN bit here in TCR if you want to */
			eth_info.total_tx_packets++;
			return;
		}
		/* Check if any error bits were set */
		if( temp & (EPHSR_TXUNRN|EPHSR_16COL|EPHSR_LATCOL) )
		{
			break;			/* Deal with ERROR */
		}
		if( SMSC_READ(TCR_REG) & (TCR_MON_CSN|TCR_STP_SQET) )
		{
			break;			/* Deal with ERROR */
		}
	}

	/* If this far, then an error has occured */
	/* Put failed packet number in PNR register so we can free memory */
//	BANK_SELECT(2);  
//	temp = (SMSC_READ(FIFO_REG) & 0x003F);		/* get current FIFO packet number */
//	SMSC_WRITE(REG_PNR,temp);					/* set that packet as accessible */
	SMSC_WRITE(MMUCR_REG,MMUCR_FREEPKT);		/* Write RELEASE SPECIFIC PACKET command (Release packet in PNR register) */

	while ( SMSC_READ(MMUCR_REG) & MMUCR_BUSY );		/* wait for MMU to become not busy */

}


/* delay function */
static void delay(int ms_delay)
{
	int i;

	i = get_timer();
	init_timer(i, ms_delay/10 );		/* Set timer in 10ms intervals */
	while( check_timer(i) );			/* Wait for timer to expire */
	free_timer( i );

}

static UINT16 PhyAccess(UINT8 write, UINT8 addr, UINT16 data)
{
	UINT8 bit_stream[64], i;
	UINT8 * stream_ptr = bit_stream;
	UINT16 ret_data = 0;

	/* start off by placing the MMI Register in it's default state */
	/* while we create the bit stream to send it */
	BANK_SELECT(3);
	SMSC_WRITE(MII_REG,0);

	/* send IDLE sequence of 32 1's */
	for(i=0;i<32;i++)
		*stream_ptr++ = MII_MDOE | MII_MDO;

	/* send start sequence of 0,1 */
	*stream_ptr++ = MII_MDOE;
	*stream_ptr++ = MII_MDOE | MII_MDO;

	if( write )
	{
		/* send write command */
		*stream_ptr++ = MII_MDOE;				/* Read Cycle - NO */ 
		*stream_ptr++ = MII_MDOE | MII_MDO;		/* Write Cyle - YES */
	}
	else
	{
		/* send read command */
		*stream_ptr++ = MII_MDOE | MII_MDO;		/* Read Cycle - YES */ 
		*stream_ptr++ = MII_MDOE;				/* Write Cyle - NO */
	}

	/* send PHY address (5 bits, mbs) */
	for(i=0;i<5;i++)
	{
		*stream_ptr++ = MII_MDOE;		/* PHY 0 address */
	}

	/* send REGISTER address (5 bits, mbs) */
	for(i=0;i<5;i++)
	{
		// shift next bit into the 6th-bit (bit 5) spot so we can check it every time
		addr <<= 1;
		if(addr & 0x20)		// make bit 5
			*stream_ptr++ = MII_MDOE | MII_MDO;
		else
			*stream_ptr++ = MII_MDOE;
	}

	/* turn around time ( need for when reading from PHY) */
	if( write )
	{
		/* output a 1, then a 0 */
		*stream_ptr++ = MII_MDOE | MII_MDO;
		*stream_ptr++ = MII_MDOE;
	}
	else
	{
		/* set to input mode */
		*stream_ptr++ = 0;
		*stream_ptr++ = 0;
	}

	/* format the write data bit if needed */
	if(write)
	{
		for(i=0;i<16;i++)
		{
			if(data & 0x8000)
				*stream_ptr++ = MII_MDOE | MII_MDO;
			else
				*stream_ptr++ = MII_MDOE;
			data <<= 1;
		}
	}


	/* now go ahead and send the created bit stream */
	stream_ptr = bit_stream;
	/* send first 48 bits (all but data section) */
	for(i=0; i<48; i++)
	{
		SMSC_WRITE(MII_REG,*stream_ptr);			// DATA with CLK LOW
		delay(0);
		SMSC_WRITE(MII_REG,*stream_ptr++ | MII_MCLK);	// DATA with CLK HIGH (PHY will latch data) (and advance pointer)
		delay(0);
	}

	/* if writting, send the rest */
	if(write)
	{
		for(i=0; i<16; i++)
		{
			SMSC_WRITE(MII_REG,*stream_ptr);			// DATA with CLK LOW
			delay(0);
			SMSC_WRITE(MII_REG,*stream_ptr++ | MII_MCLK);	// DATA with CLK HIGH (PHY will latch data) (and advance pointer)
			delay(0);
		}
	}
	else
	{
		/* read data using MDI pin in register */
		for(i=0; i<16; i++)
		{
			ret_data <<= 1;

			SMSC_WRITE(MII_REG,0);			// CLK pin LOW (PHY will put next bit on the line)
			delay(0);
			ret_data += ((SMSC_READ(MII_REG) & MII_MDI) / 2);	/* read input pin, mask and shift */
			SMSC_WRITE(MII_REG,MII_MCLK);	// CLK pin HIGH (tell PHY to give us the next bit)
			delay(0);
		}
	}

	SMSC_WRITE(MII_REG,0);		/* Put register back to idle state */
	return ret_data;
}


static void SMsC_Write(unsigned int offset, unsigned int data)
{
	*( (_far unsigned int *)(BASE_ADDR + offset) ) = data;
}
static unsigned int SMsC_Read(unsigned int offset)
{
	return *((volatile _far unsigned int *)(BASE_ADDR + offset));
} 

#if (INTERRUPT_ETHERNET)
#pragma INTERRUPT SmscIntr
void SmscIntr(void)
{
	UINT16 saved_bank,status;

	/* save current bank selection */
	saved_bank = SMSC_READ(BSR_REG);

	/* Check to see what type of interrupt it is */
	BANK_SELECT(2);   
	status = SMSC_READ(IST_REG);	/* Read Interrupt Status Register */

	if(status & IST_RCV_INT)
	{
		rx_buffer_filled = 3;		/* signal that we want the buffer to get filled */
		NE2000ReceiveFrame();
	}

	/* Check if a receive error has occured */
	if( status & IST_RX_OVRN_INT)
	{
		/* This bit will stay latched untill reset */
		SMSC_WRITE(ACK_REG,IST_RX_OVRN_INT);
//		SMSC_WRITE(MMUCR_REG,MMUCR_RESET);
//		while ( SMSC_READ(MMUCR_REG) & MMUCR_BUSY );		/* wait for MMU to become not busy */
	}

	/* put bank register back the way it was */
	SMSC_WRITE(BSR_REG, saved_bank);
}
#endif

#if (POLLED_ETHERNET)
#pragma INTERRUPT SmscIntr
void SmscIntr(void)
{
	// dummy function to satisfy the vector table entry
}
#endif