/* ---------------------------------------------------------------------------
          Copyright (c) 2004-2006 Micrel, Inc.  All rights reserved.
   ---------------------------------------------------------------------------

    transmit.c - Linux network device driver transmit processing.

    Author      Date        Description
    THa         01/05/04    Created file.
    THa         06/27/05    Updated for version 0.1.5.
    THa         10/18/05    Updated for Renesas 7751R board.
    THa         01/13/06    Transfer dword-aligned data for performance.
    THa         03/15/06    Correct 10Mbit transmit stop problem.
    THa         04/03/06    Use TWO_NETWORK_INTERFACE instead of TWO_PORTS.
    THa         06/05/06    Make sure interrupts are enabled before processing
                            them.
    THa         06/29/06    Change statistics reporting to provide meaningful
                            data.
    THa         07/26/06    Add preliminary DMA support.
    THa         10/16/06    Add ST7100 DIM_0_x_SG FDMA support.
    THa         12/05/06    Version 1.0.8 adds revision A6 turbo mode support.
   ---------------------------------------------------------------------------
*/


#include "target.h"
#include "hardware.h"
#include "device.h"
#include <asm/irq.h>

#ifdef SH_BUS_DMA
#include <linux/delay.h>

#ifdef CONFIG_STM_DMA
#include <asm/dma.h>

extern struct dma_channel* stm_dma[];
extern int stm_dma_chan[];
#endif

#ifdef CONFIG_SH_7751_SOLUTION_ENGINE
BOOLEAN Platform_DmaStartXfer ( PHARDWARE, UCHAR, ULONG, ULONG, ULONG, BOOLEAN );
BOOLEAN Platform_IsDmaComplete ( PHARDWARE, UCHAR );
#endif

extern unsigned char* dma_src;
extern unsigned char* dma_dst;
#endif

#ifdef DBG
#if 0
#define DEBUG_TX_DATA
#endif
#if 0
#define SIMULATE_TIMEOUT
#define SEND_STOP_COUNT  3000
#endif
#endif

#if !defined( SH_BUS_DMA ) &&  !defined( NO_TX_INT )
#if 1
#define SKB_WAITING
#endif
#endif

/* -------------------------------------------------------------------------- */

#ifdef SH_BUS_DMA
struct sk_buff* dma_tx_skb = NULL;

int dma_send_packet (
    struct sk_buff*    skb,
    struct net_device* dev )
{
    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;
    int              len;
    char*            data;

#if defined( TWO_NETWORK_INTERFACE )  ||  !defined( NO_STATS )
    int              port = MAIN_PORT;
#endif

#ifdef TWO_NETWORK_INTERFACE
    /* Indicate which port has sent packet. */
    port = priv->port;
    pHardware->m_bPortTX = DEV_TO_HW_PORT( port );
    pHardware->m_fPortTX |= port + 1;
#endif

#if ETH_ZLEN > 60
    len = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;
#else
    /* Hardware will pad the length to 60. */
    len = skb->len;
#endif
    data = skb->data;

    priv->skb_waiting = NULL;
    HardwareSetTransmitLength( pHardware, len );

    priv->stats.tx_packets++;
    priv->stats.tx_bytes += len;

    dev->trans_start = jiffies;

#ifdef SH_32BIT_ALIGNED
    memcpy( pHardware->m_bLookahead, data, len );
    data = pHardware->m_bLookahead;
#endif

    len = ( len + 3 ) & ~3;
    dma_cache_wback( data, len );

    dma_tx_skb = skb;

#ifdef CONFIG_STM_DMA
    stm_dma[ 1 ]->dar = virt_to_bus(( void* )( 
        pHardware->m_ulVIoAddr + REG_DATA_OFFSET ));
    stm_dma[ 1 ]->flags = NODE_DONE_ISR | DIM_1_x_0 | LIST_TYPE_UNLINKED;

#ifdef SH_BUS_DMA_PACED
    stm_dma[ 1 ]->flags |= 13;
#endif
    stm_dma[ 1 ]->count = len;
    stm_dma[ 1 ]->sar = virt_to_bus( data );
    dma_configure_channel( stm_dma[ 1 ]);
    dma_xfer( stm_dma_chan[ 1 ]);
#endif

#ifdef CONFIG_SH_7751_SOLUTION_ENGINE
    Platform_DmaStartXfer( pHardware, 1,
        virt_to_bus( data ),
        pHardware->m_ulVIoAddr + REG_DATA_OFFSET, 
        len, TRUE );
#endif
    return 0;
}  /* dma_send_packet */


void dma_send_packet_done (
    struct net_device* dev )
{
    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;

    HW_WRITE_BYTE( pHardware, REG_TXQ_CMD_OFFSET, TXQ_CMD_ENQUEUE_PACKET );

    dev_kfree_skb_irq( dma_tx_skb );
    dma_tx_skb = NULL;

    ReleaseHardware( hw_priv, TRUE );
}  /* dma_send_packet_done */
#endif

/* -------------------------------------------------------------------------- */

#ifdef SIMULATE_TIMEOUT
static int sendcount = 0;
#endif

/*
    send_packet

    Description:
        This function is used to send a packet out to the network.

    Parameters:
        struct sk_buff* skb
            Pointer to socket buffer.

        struct net_device* dev
            Pointer to network device.

    Return (int):
        Zero if successful; otherwise an error code indicating failure.
*/

static int send_packet (
    struct sk_buff*    skb,
    struct net_device* dev )
{
    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;
    int              len;

#if defined( TWO_NETWORK_INTERFACE )  ||  !defined( NO_STATS )
    int              port = MAIN_PORT;
#endif
    int              rc = 1;

#ifdef TWO_NETWORK_INTERFACE
    /* Indicate which port has sent packet. */
    port = priv->port;
    pHardware->m_bPortTX = DEV_TO_HW_PORT( port );
    pHardware->m_fPortTX |= port + 1;
#endif

#if ETH_ZLEN > 60
    len = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;
#else
    /* Hardware will pad the length to 60. */
    len = skb->len;
#endif

    HardwareSetTransmitLength( pHardware, len );

#ifdef SH_32BIT_ALIGNED
    memcpy( pHardware->m_bLookahead, skb->data, len );
    HW_WRITE_BUFFER( pHardware, REG_DATA_OFFSET, pHardware->m_bLookahead,
        len );

#else
    HW_WRITE_BUFFER( pHardware, REG_DATA_OFFSET, skb->data, len );
#endif
    HW_WRITE_BYTE( pHardware, REG_TXQ_CMD_OFFSET, TXQ_CMD_ENQUEUE_PACKET );

#if 0
    if ( HardwareSendPacket( pHardware ) )
#endif
    {
        rc = 0;

        /* Update transmit statistics. */
#ifndef NO_STATS
        pHardware->m_cnCounter[ port ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;
        pHardware->m_cnCounter[ port ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] +=
            len;
#endif
        priv->stats.tx_packets++;
        priv->stats.tx_bytes += len;

        dev->trans_start = jiffies;

#ifdef SIMULATE_TIMEOUT
        ++sendcount;
        if ( SEND_STOP_COUNT == sendcount  ) {
            DBG_PRINT( "stop sending:%lu\n", jiffies );
            netif_stop_queue( dev );
        }
#endif
    }

#ifdef SKB_WAITING
    priv->skb_waiting = NULL;
#endif

#ifdef DEBUG_TX_DATA
    {
        int    nLength;
        UCHAR* bData = ( PUCHAR ) skb->data;

        for ( nLength = 0; nLength < len; nLength++ )
        {
            DBG_PRINT( "%02x ", bData[ nLength ]);
            if ( ( nLength % 16 ) == 15 )
            {
                DBG_PRINT( NEWLINE );
            }
        }
        DBG_PRINT( NEWLINE );
    }
#endif

    return( rc );
}  /* send_packet */


#ifdef SKB_WAITING
/*
    send_next_packet

    Description:
        This function is used to send the next packet waiting to be sent.

    Parameters:
        struct net_device* dev
            Pointer to network device.

    Return (int):
        Zero if successful; otherwise an error code indicating failure.
*/

static int send_next_packet (
    struct net_device* dev )
{
    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;
    int              rc = 1;

    if ( priv->skb_waiting  &&
            HardwareAllocPacket( pHardware, priv->skb_waiting->len ) ) {
        struct sk_buff* skb = priv->skb_waiting;

        rc = send_packet( priv->skb_waiting, dev );
        dev_kfree_skb_irq( skb );
        netif_wake_queue( dev );
    }
    return( rc );
}  /* send_next_packet */
#endif


/*
    transmit_done

    Description:
        This routine is called when the transmit interrupt is triggered,
        indicating either a packet is sent successfully or there are transmit
        errors.

    Parameters:
        struct net_device* dev
            Pointer to network device.

    Return (None):
*/

void
ks8842i_\
transmit_done (
    struct net_device* dev )
{
    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;

#ifdef TWO_NETWORK_INTERFACE
    int              port = MAIN_PORT;

    port = HW_TO_DEV_PORT( pHardware->m_bPortTX );
    if ( OTHER_PORT == port ) {
        dev = hw_priv->pDev[ OTHER_PORT ];
        priv = ( struct dev_priv* ) dev->priv;
    }
#endif
    if ( !HardwareTransmitDone( pHardware ) ) {

        /* Update transmit error statistics. */
    }

#ifdef TWO_NETWORK_INTERFACE
    dev = hw_priv->pDev[ MAIN_PORT ];
#endif

#ifdef SIMULATE_TIMEOUT
    if ( sendcount < SEND_STOP_COUNT )
#endif
    if ( netif_queue_stopped( dev ) ) {

#ifdef SKB_WAITING
        send_next_packet( dev );

#else
#ifdef SH_BUS_DMA
        priv->dev_state &= ~DEV_STATE_STOP_TX_BY_TX;
        if ( !( priv->dev_state & DEV_STATE_STOP_TX_BY_RX ) )
#endif
        netif_wake_queue( dev );
#endif
    }

#ifdef TWO_NETWORK_INTERFACE
    dev = hw_priv->pDev[ OTHER_PORT ];
    if ( netif_queue_stopped( dev ) ) {

#ifdef SKB_WAITING
        send_next_packet( dev );

#else
#ifdef SH_BUS_DMA
        priv->dev_state &= ~DEV_STATE_STOP_TX_BY_TX;
        if ( !( priv->dev_state & DEV_STATE_STOP_TX_BY_RX ) )
#endif
        netif_wake_queue( dev );
#endif
    }
#endif
}  /* transmit_done */


/*
    transmit_empty

    Description:
        This routine is called when the transmit empty interrupt is triggered,
        indicating no more packets to be sent.

    Parameters:
        struct net_device* dev
            Pointer to network device.

    Return (None):
*/

void
ks8842i_\
transmit_empty (
    struct net_device* dev )
{

#ifdef TWO_NETWORK_INTERFACE
    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;
    int              port = MAIN_PORT;

    for ( port = 0; port < 2; port++ ) {
        if ( OTHER_PORT == port ) {
            dev = hw_priv->pDev[ OTHER_PORT ];
            priv = ( struct dev_priv* ) dev->priv;
        }

        /* This port has sent packets. */
        if ( ( pHardware->m_fPortTX & ( port + 1 )) ) {
#endif
            if ( netif_queue_stopped( dev ) )
                netif_wake_queue( dev );

#ifdef TWO_NETWORK_INTERFACE
        }
#endif

#ifdef SKB_WAITING
        send_next_packet( dev );
#endif

#ifdef TWO_NETWORK_INTERFACE
    }
    pHardware->m_fPortTX = 0;
#endif
}  /* transmit_empty */

/* -------------------------------------------------------------------------- */

/*
    dev_transmit

    Description:
        This function is used by the upper network layer to send out a packet.

    Parameters:
        struct sk_buff* skb
            Pointer to socket buffer.

        struct net_device* dev
            Pointer to network device.

    Return (int):
        Zero if successful; otherwise an error code indicating failure.
*/

int
ks8842i_\
dev_transmit (
    struct sk_buff*    skb,
    struct net_device* dev )
{
    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;
    int              len;
    int              rc = 0;

#ifndef SH_BUS_DMA
#ifdef TEST_BLOCK_INTERRUPT
    UINT             InterruptMask;
#endif
#endif

#ifdef SKB_WAITING
    /* There is a packet waiting to be sent.  Transmit queue should already be
       stopped.
    */
    if ( priv->skb_waiting ) {

#ifdef DBG
        printk( "waiting to send!\n" );
#endif
        priv->stats.tx_dropped++;
        if ( !netif_queue_stopped( dev ) ) {
            DBG_PRINT( "queue not stopped!\n" );
            netif_stop_queue( dev );
        }
        return 1;
    }
#endif

#if ETH_ZLEN > 60
    len = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;
#else
    /* Hardware will pad the length to 60. */
    len = skb->len;
#endif

    if ( ( rc = AcquireHardware( hw_priv, FALSE, FALSE )) ) {
        return( rc );
    }

#ifndef SH_BUS_DMA
#ifdef TEST_BLOCK_INTERRUPT

    /* Save the current interrupt mask and block all interrupts. */
    InterruptMask = HardwareBlockInterrupt( pHardware );

#else
    spin_lock_irq( &priv->lock );
#endif
#endif

#ifdef SKB_WAITING
    priv->skb_waiting = skb;
#endif
    if ( HardwareAllocPacket( pHardware, len ) ) {

#ifndef SH_BUS_DMA_WRITE
        rc = send_packet( skb, dev );
        dev_kfree_skb( skb );

#else
        rc = dma_send_packet( skb, dev );
#endif
    }
    else {

#ifdef DBG_
        printk( "wait to send:%lu\n", jiffies );
#endif

/* Transmit interrupt is disabled, so network queue cannot be restarted. */
#ifndef NO_TX_INT
        /* Stop the transmit queue until packet is allocated. */
#ifdef SH_BUS_DMA
        if ( !( priv->dev_state &
                ( DEV_STATE_STOP_TX_BY_RX | DEV_STATE_STOP_TX_BY_TX )) )
#endif
        netif_stop_queue( dev );
#ifdef SH_BUS_DMA
        priv->dev_state |= DEV_STATE_STOP_TX_BY_TX;
#endif
#endif

#ifndef SKB_WAITING
        rc = -ENOMEM;
#endif
    }

#ifndef SH_BUS_DMA
#ifdef TEST_BLOCK_INTERRUPT

    /* Restore the interrupt mask. */
    HardwareRestoreInterrupt( pHardware, InterruptMask );

#else
    spin_unlock_irq( &priv->lock );
#endif
#endif

#ifndef SH_BUS_DMA_WRITE
    ReleaseHardware( hw_priv, FALSE );
#endif

    return( rc );
}  /* dev_transmit */

/* -------------------------------------------------------------------------- */

/*
    dev_transmit_timeout

    Description:
        This routine is called when the transmit timer expires.  That indicates
        the hardware is not running correctly because transmit interrupts are
        not triggered to free up resources so that the transmit routine can
        continue sending out packets.  The hardware is reset to correct the
        problem.

    Parameters:
        struct net_device* dev
            Pointer to network device.

    Return (None):
*/

void
ks8842i_\
dev_transmit_timeout (
    struct net_device *dev )
{
    static unsigned long last_reset = 0;

    struct dev_priv* priv = ( struct dev_priv* ) dev->priv;
    struct dev_info* hw_priv = priv->pDevInfo;
    PHARDWARE        pHardware = &hw_priv->hw;
    int              port;

#ifdef DBG
    printk( "transmit timeout:%lu\n", jiffies );
#endif

#ifdef TWO_NETWORK_INTERFACE

    /* Only reset the hardware if time between calls is long. */
    if ( jiffies - last_reset <= dev->watchdog_timeo )
        hw_priv = NULL;
#endif
    last_reset = jiffies;
    if ( hw_priv ) {
        AcquireHardware( hw_priv, FALSE, FALSE );

        HardwareDisableInterrupt( pHardware );
        HardwareDisable( pHardware );

#ifndef SIMULATE_TIMEOUT
        HardwareReset( pHardware );
#endif
        HardwareSetup( pHardware );
            
#ifdef TWO_NETWORK_INTERFACE
        PortSet_STP_State( pHardware, MAIN_PORT, STP_STATE_BLOCKED );
        PortSet_STP_State( pHardware, OTHER_PORT, STP_STATE_BLOCKED );
#endif

        /* Reset may wipe out these registers. */
        if ( pHardware->m_bMulticastListSize )
            HardwareSetGroupAddress( pHardware );
    }

#ifdef TWO_NETWORK_INTERFACE
    for ( port = MAIN_PORT; port <= OTHER_PORT; port++ ) {
        struct net_device* net_if = priv->pDevInfo->pDev[ port ];

        /* The second net device is not created. */
        if ( !net_if )
            continue;

        priv = ( struct dev_priv* ) net_if->priv;

#else
    for ( port = MAIN_PORT; port < OTHER_PORT; port++ ) {
#endif

        /* This port device is opened for use. */
        if ( priv->opened ) {
        
#ifdef TWO_NETWORK_INTERFACE
            PortSet_STP_State( pHardware, port, STP_STATE_FORWARDING );
#endif

#ifdef SKB_WAITING
            if ( priv->skb_waiting ) {
                priv->stats.tx_dropped++;
                dev_kfree_skb( priv->skb_waiting );
                priv->skb_waiting = NULL;
            }
#endif
        }
    }

    dev->trans_start = jiffies;
    netif_wake_queue( dev );

    if ( hw_priv ) {
        HardwareEnable( pHardware );

        ReleaseHardware( hw_priv, FALSE );
        HardwareEnableInterrupt( pHardware );
    }

#ifdef SIMULATE_TIMEOUT
    sendcount = 0;
#endif
}  /* dev_transmit_timeout */
